#include "framebuffer.h"
#include "math.h"
#include <iostream>
#include "scene.h"
#include "v3.h"
#include <iostream>

#include "tiffio.h"

using namespace std;

FrameBuffer::FrameBuffer(int u0, int v0,int _w, int _h) 
	: Fl_Gl_Window(u0, v0, _w, _h, 0) {

	ishw = false;
	isgpu = false;
	w = _w;
	h = _h;
	pix = new unsigned int[w*h];
	zb = new float[w*h];

}


void FrameBuffer::draw() {

	if (ishw) {
		scene->RenderHW();
	}
	else if (isgpu) {
		scene->RenderGPU();
	}
	else {
		glDrawPixels(w, h, GL_RGBA, GL_UNSIGNED_BYTE, pix);
	}

}

int FrameBuffer::handle(int event) {

	switch (event)
	{
	case FL_KEYBOARD: {
		KeyboardHandle();
		return 0;
	}
	default:
		break;
	}
	return 0;
}

void FrameBuffer::KeyboardHandle() {

	int key = Fl::event_key();
	switch (key) {
	case ',': {
		cerr << "INFO: pressed ," << endl;
		break;
	}
	default:
		cerr << "INFO: do not understand keypress" << endl;
	}
}


void FrameBuffer::SetGuarded(int u, int v, unsigned int color) {

	if (u < 0 || u > w - 1 || v < 0 || v > h - 1)
		return;

	Set(u, v, color);

}


void FrameBuffer::Set(int u, int v, unsigned int color) {

	pix[(h - 1 - v)*w + u] = color;

}

void FrameBuffer::SetBGR(unsigned int bgr) {

	for (int uv = 0; uv < w*h; uv++)
		pix[uv] = bgr;

}


// load a tiff image to pixel buffer
void FrameBuffer::LoadTiff(char* fname) {
	TIFF* in = TIFFOpen(fname, "r");
	if (in == NULL) {
		cerr << fname << " could not be opened" << endl;
		return;
	}

	int width, height;
	TIFFGetField(in, TIFFTAG_IMAGEWIDTH, &width);
	TIFFGetField(in, TIFFTAG_IMAGELENGTH, &height);
	if (w != width || h != height) {
		w = width;
		h = height;
		delete[] pix;
		pix = new unsigned int[w*h];
		size(w, h);
		glFlush();
		glFlush();
	}

	if (TIFFReadRGBAImage(in, w, h, pix, 0) == 0) {
		cerr << "failed to load " << fname << endl;
	}

	TIFFClose(in);
}

// save as tiff image
void FrameBuffer::SaveAsTiff(char *fname) {

	TIFF* out = TIFFOpen(fname, "w");

	if (out == NULL) {
		cerr << fname << " could not be opened" << endl;
		return;
	}

	TIFFSetField(out, TIFFTAG_IMAGEWIDTH, w);
	TIFFSetField(out, TIFFTAG_IMAGELENGTH, h);
	TIFFSetField(out, TIFFTAG_SAMPLESPERPIXEL, 4);
	TIFFSetField(out, TIFFTAG_BITSPERSAMPLE, 8);
	TIFFSetField(out, TIFFTAG_ORIENTATION, ORIENTATION_TOPLEFT);
	TIFFSetField(out, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG);
	TIFFSetField(out, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_RGB);

	for (uint32 row = 0; row < (unsigned int)h; row++) {
		TIFFWriteScanline(out, &pix[(h - row - 1) * w], row);
	}

	TIFFClose(out);
}


void FrameBuffer::DrawRectangle(int u0, int v0, int u1, int v1, unsigned int color) {

	if (!ClipToScreen(u0, v0, u1, v1))
		return;

	for (int u = u0; u <= u1; u++) {
		for (int v = v0; v <= v1; v++) {
			Set(u, v, color);
		}
	}

}

void FrameBuffer::DrawCircle(int u0, int v0, int r, unsigned int color) {

	int left = u0 - r;
	int right = u0 + r;
	int top = v0 - r;
	int bottom = v0 + r;
	if (!ClipToScreen(left, right, top, bottom))
		return;

	V3 cv(.5f + (float)u0, .5f + (float)v0, 0.0f);
	for (int u = left; u <= right; u++) {
		for (int v = top; v <= bottom; v++) {
			V3 pv(.5f + (float)u, .5f + (float)v, 0.0f);
			V3 dv = pv - cv;
			if (dv*dv > r*r)
				continue;
			Set(u, v, color);
		}
	}

}

int FrameBuffer::ClipToScreen(int &u0, int &v0, int &u1, int &v1) {


	if (u1 < 0)
		return 0;
	if (u0 > w - 1)
		return 0;
	if (v1 < 0)
		return 0;
	if (v0 > h - 1)
		return 0;

	if (u0 < 0)
		u0 = 0;
	if (v0 < 0)
		v0 = 0;
	if (u1 > w - 1)
		u1 = w - 1;
	if (v1 > h - 1)
		v1 = h - 1;
	return 1;

}

void FrameBuffer::DrawSegment(V3 p0, V3 c0, V3 p1, V3 c1) {

	V3 p1p0 = p1 - p0;
	V3 c1c0 = c1 - c0;
	int pixN;
	if (fabsf(p1p0[0]) > fabsf(p1p0[1])) {
		pixN = (int)(fabsf(p1p0[0]) + 1);
	}
	else {
		pixN = (int)(fabsf(p1p0[1]) + 1);
	}
	for (int stepsi = 0; stepsi < pixN+1; stepsi++) {
		float fracf = (float) stepsi / (float) (pixN);
		V3 p = p0 + p1p0 * fracf;
		V3 c = c0 + c1c0 * fracf;
		int u = (int) p[0];
		int v = (int) p[1];
		if (u < 0 || u > w - 1 || v < 0 || v > h - 1)
			continue;
		if (!Visible(u, v, p[2]))
			continue;
		Set(u, v, c.GetColor());
	}

}

void FrameBuffer::ClearZB(float z0) {

	for (int i = 0; i < w*h; i++)
		zb[i] = z0;

}


int FrameBuffer::Visible(int u, int v, float currz) {

	int uv = (h - 1 - v)*w + u;
	if (zb[uv] > currz)
		return 0;
	zb[uv] = currz;
	return 1;

}

void FrameBuffer::Draw3DPoint(V3 p, V3 c, PPC *ppc, int psize) {

	V3 pp;
	if (!ppc->Project(p, pp))
		return;

	if (pp[0] < 0.0f || pp[0] >= (float)w || pp[1] < 0.0f || pp[1] >= (float)h)
		return;

	if (!Visible(pp[0], pp[1], pp[2]))
		return;

	int u = (int)pp[0];
	int v = (int)pp[1];
	DrawRectangle(u - psize / 2, v - psize / 2, u + psize / 2, v + psize / 2, c.GetColor());

}

void FrameBuffer::Draw3DSegment(V3 p0, V3 c0, V3 p1, V3 c1, PPC *ppc) {

	V3 pp0, pp1;
	if (!ppc->Project(p0, pp0))
		return;
	if (!ppc->Project(p1, pp1))
		return;

	DrawSegment(pp0, c0, pp1, c1);

}


void FrameBuffer::Visualize(PPC *ppc0, float currf, PPC *ppc1, FrameBuffer *fb1) {

	for (int v = 0; v < h; v++) {
		for (int u = 0; u < w; u++) {
			if (zb[(h - 1 - v)*w + u] == 0.0f)
				continue;
			V3 pixP = ppc0->UnprojectPixel(.5f + (float)u, .5f + (float)v, currf);
			V3 cv; cv.SetFromColor(Get(u, v));
			fb1->Draw3DPoint(pixP, cv, ppc1, 1);
		}
	}

}

void FrameBuffer::Visualize3D(PPC *ppc0, float currf, PPC *ppc1, FrameBuffer *fb1) {

	for (int v = 0; v < h; v++) {
		for (int u = 0; u < w; u++) {
			if (zb[(h - 1 - v)*w + u] == 0.0f)
				continue;
			V3 pP(.5f + (float)u, .5f + (float)v, zb[(h - 1 - v)*w + u]);
			V3 pixP = ppc0->Unproject(pP);
			V3 cv; cv.SetFromColor(Get(u, v));
			fb1->Draw3DPoint(pixP, cv, ppc1, 1);
		}
	}

}


unsigned int FrameBuffer::GetGuarded(int u, int v) {

	if (u < 0 || u > w - 1 || v < 0 || v > h - 1)
		return 0xFFFFFF00;

	return Get(u, v);

}



unsigned int FrameBuffer::Get(int u, int v) {


	return pix[(h - 1 - v)*w + u];

}


void FrameBuffer::Resample(PPC *ppc0, FrameBuffer *fb1, PPC *ppc1) {

	for (int v = 0; v < h; v++) {
		for (int u = 0; u < w; u++) {
			float uf = .5f + (float)u;
			float vf = .5f + (float)v;
			V3 P = ppc0->UnprojectPixel(uf, vf, 100.0f);
			V3 PP;
			if (!ppc1->Project(P, PP))
				continue;
			unsigned int c1 = fb1->GetGuarded((int) PP[0], (int) PP[1]);
			if (c1 == 0xFFFFFF00)
				continue;
			V3 c1v; c1v.SetFromColor(c1);
			unsigned int c0 = Get(u, v);
			if (c0 != 0xFF000000) {
				V3 c0v; c0v.SetFromColor(c0);
				V3 cv = (c1v + c0v) * 0.5f;
				c1 = cv.GetColor();
			}
			Set(u, v, c1);
		}
	}

}

float FrameBuffer::ColorDifference(PPC *ppc0, FrameBuffer *fb1, PPC *ppc1) {

	float ret = 0.0f;
	int opn = 0;
	for (int v = 0; v < h; v++) {
		for (int u = 0; u < w; u++) {
			float uf = .5f + (float)u;
			float vf = .5f + (float)v;
			V3 P = ppc0->UnprojectPixel(uf, vf, 100.0f);
			V3 PP;
			if (!ppc1->Project(P, PP))
				continue;
			unsigned int c1 = fb1->GetGuarded((int)PP[0], (int)PP[1]);
			if (c1 == 0xFFFFFF00)
				continue;
			V3 c1v; c1v.SetFromColor(c1);
			unsigned int c0 = Get(u, v);
			V3 c0v; c0v.SetFromColor(c0);
			ret += (c1v - c0v)*(c1v-c0v);
			opn++;
		}
	}

	ret /= opn*3;
	ret = sqrtf(ret);
	ret *= 255.0f;
	return ret;

}

